@@ -1,4 +1,4 @@
-from flask.views import MethodView
+from flask.views import MethodView, MethodViewType
 from flask import jsonify, current_app, request
 
 import json
@@ -7,6 +7,7 @@ import sqlalchemy
 import sys
 import calendar
 import datetime
+import traceback
 
 
 class LeverException(Exception):
@@ -95,13 +96,101 @@ OPERATORS = {
 }
 
 
-class API(MethodView):
+class preprocess(object):
+    """ Simple decorator that sets special attributes on decorated methods.
+    These attributes get picked up by our __new__ method and assign
+    preprocessors for specific actions or methods """
+    def __init__(self, method=None, action=None):
+        self.action = action
+        self.method = method
+
+    def __call__(self, f):
+        f._pre_action = self.action
+        f._pre_method = self.method
+        return f
+
+class postprocess(preprocess):
+    def __call__(self, f):
+        f._post_action = self.action
+        f._post_method = self.method
+        return f
+
+
+class ImpersonateMixin(object):
+    """ Allows ACL lookups from a user object. Expects current_user attribute
+    to be set to an application LocalProxy that yeilds the user object for the
+    request. This is the most common ACL system used with Lever. """
+    from flask.ext.login import current_user
+    user_model = None
+
+    @preprocess('post')
+    def impersonate(self):
+        if self.user_model is None:
+            raise Exception("user_model must be defined to lookup users")
+        # check to ensure the user can create for others if requested
+        username = self.params.pop('__username', None)
+        userid = self.params.pop('__user_id', None)
+        if userid or username:
+            query = self.session.query(self.user_model)
+            if userid:
+                query = query.filter_by(id=userid)
+            if username:
+                query = query.filter_by(username=username)
+            user = query.one()
+            assert self.can_cls('class_create_other'), "Cant create for other users"
+            self.params['user'] = user
+        else:
+            assert self.can_cls('class_create'), "Cant create that object"
+            self.params['user'] = self.current_user.get()
+
+    def can_cls(self, action):
+        # don't pass in impersonated user objects...
+        dct = dict((k, v) for k, v in self.params.items() if k != 'user')
+        return self.model.can_cls(action, **dct)
+
+
+class ModelBasedACL(object):
+    """ Executes ACL methods on the model directly to determine ability to
+    execute requested action """
+    def can(self, obj, action):
+        return obj.can(action)
+
+    def can_cls(self, action):
+        return self.model.can_cls(action, **self.params)
+
+
+class APIMeta(MethodViewType):
+    def __init__(mcs, name, bases, dct):
+        MethodViewType.__init__(mcs, name, bases, dct)
+        types = ['_pre_method', '_post_method', '_pre_action', '_post_action']
+        for key in types:
+            setattr(mcs, key, {})
+        attrs = list(six.itervalues(dct))
+        for base in bases:
+            attrs.extend(list(six.itervalues(base.__dict__)))
+        for attr in attrs:
+            for key in types:
+                val = getattr(attr, key, None)
+                if val:
+                    if not isinstance(val, (list, tuple)):
+                        val = (val, )
+                    for method in val:
+                        getattr(mcs, key).setdefault(method, []).append(attr)
+
+
+class API(six.with_metaclass(APIMeta, MethodView)):
+    """ The main method that underlies Lever. Create a new class that inherits
+    from this and set the session to your database session, model to the model
+    that you're wrapping and basic functionality should be good to go. """
     max_pg_size = 100
+    # Defines the maximum pagination size that the client can select
     pkey_val = 'id'
-    create_method = 'create'
+    # defines the primary key for your model. this value will be expected on
+    # get and updates
+    create_method = '__init__'
     params = {}
     session = None
-    current_user = None
+    # The database session from SQLAlchemy
 
     @property
     def pkey(self):
@@ -128,13 +217,14 @@ class API(MethodView):
 
         extra = {'original_exc': str(info[1]),
                  'original_exc_type': str(info[0]),
+                 'tb': "".join(traceback.format_tb(info[2])),
                  'params': self.params}
         end_user = {'success': False}
         code = None
 
         try:
             # reraise to handle differnet exceptions differently
-            raise info[1]
+            six.reraise(*info)
         # Common API errors
         except KeyError as e:
             msg = 'Incorrect syntax or missing key ' + str(e)
@@ -153,7 +243,7 @@ class API(MethodView):
             extra.update(e.extra)
             e.extra = extra
             e.end_user = end_user
-            raise e
+            six.reraise(LeverException, e, tb=sys.exc_info()[2])
 
         # SQLAlchemy exceptions
         except sqlalchemy.orm.exc.NoResultFound:
@@ -184,24 +274,32 @@ class API(MethodView):
     def get_obj(self):
         pkey = self.params.pop(self.pkey_val, None)
         if pkey:  # if a int primary key is passed
-            return self.model.query.filter(self.pkey == pkey).one()
+            return self.session.query(self.model).filter(self.pkey == pkey).one()
         return False
 
+    def can(self, obj, action):
+        """ This function should parse the current parameters to gain parent
+        information for properly running can_cls on the model this API wraps
+        """
+        return True
+
     def can_cls(self, action):
         """ This function should parse the current parameters to gain parent
         information for properly running can_cls on the model this API wraps
         """
-        return self.model.can_cls(action)
+        return True
 
     def get(self):
         """ Retrieve an object from the database """
         # convert args to a real dictionary that can be popped
         self.params = dict((one, two) for one, two in six.iteritems(request.args))
+        for method in self._pre_method.get('get', []):
+            method(self)
         join = self.params.pop('join_prof', 'standard_join')
         obj = self.get_obj()
         if obj:  # if a int primary key is passed
-            assert obj.can('view_' + join), "Can't view that object with join " + join
-            return jsonify(success=True, objects=[get_joined(obj, join)])
+            assert self.can(obj, 'view_' + join), "Can't view that object with join " + join
+            retval = dict(success=True, objects=[get_joined(obj, join)])
         else:
             query = self.search()
             one = self.params.pop('__one', None)
@@ -210,66 +308,45 @@ class API(MethodView):
             else:
                 query = self.paginate(query=query)
             for obj in query:
-                assert obj.can('view_' + join), "Can't view that object with join " + join
-            return jsonify(success=True, objects=get_joined(query, join))
+                assert self.can(obj, 'view_' + join), "Can't view that object with join " + join
+            retval = dict(success=True, objects=get_joined(query, join))
+
+        for method in self._post_method.get('get', []):
+            method(self, retval)
+        return jsonify(**retval)
 
     def post(self):
         """ Create a new object """
-        self.params = request.get_json(silent=True)
-        if not self.params:
-            raise LeverSyntaxError("To create, values must be specified")
-        # check to ensure the user can create for others if requested
-        username = self.params.pop('__username', None)
-        userid = self.params.pop('__user_id', None)
-        if userid or username:
-            query = self.model.query
-            if userid:
-                query = query.filter_by(id=userid)
-            if username:
-                query = query.filter_by(username=username)
-            self.params['user'] = query.one()
-            self.create_hook()
-            assert self.can_cls('class_create_other', params=self.params), "Cant create for other users"
-        else:
-            self.params['user'] = self.current_user.get()
-            self.create_hook()
-            assert self.can_cls('class_create'), "Cant create that object"
-
-        try:
-            model = getattr(self.model, self.create_method)(**self.params)
-        except TypeError as e:
-            if 'argument' in str(e):
-                msg = "Wrong number of arguments supplied for create."
-                six.reraise(LeverSyntaxError, msg, tb=sys.exc_info()[2])
-            else:
-                raise
-
-        self.session.commit()
-        if model:  # only return the model if we recieved it back
-            return jsonify(success=True, objects=[get_joined(model)])
-        else:
-            return jsonify(success=True)
-
-    def patch(self):
-        """ Used to execute methods on an object """
-        self.params = request.get_json(silent=True)
-        if not self.params:
-            raise LeverSyntaxError("To run an action, values must be specified")
-
-        action = self.params.pop('__action')
+        self.params = request.get_json(silent=True) or {}
+        for method in self._pre_method.get('post', []):
+            method(self)
+        action = self.params.pop('__action', None)
         cls = self.params.pop('__cls', None)
-        if cls is None:
+        if not action:
+            cls = True
+            action = self.create_method
+
+        if not cls:
             obj = self.get_obj()
             if not obj:
-                raise LeverNotFound("Could not find any object to perform an action on")
-            assert obj.can('action_' + action), "Cant perform action " + action
+                raise LeverNotFound(
+                    "Could not find any object to perform an action on")
+            assert self.can(obj, 'action_' + action), "Cant perform action " + action
         else:
             assert self.can_cls('action_' + action), "Can't perform cls action " + action
             obj = self.model
 
         retval = {}
         try:
-            ret = getattr(obj, action)(**self.params)
+            if action == '__init__':
+                obj = obj(**self.params)
+                self.session.add(obj)
+                self.session.flush()
+                ret = {'objects': [get_joined(obj)]}
+            else:
+                ret = getattr(obj, action)(**self.params)
+                if isinstance(ret, self.model):
+                    ret = {'objects': [get_joined(ret)]}
         except TypeError as e:
             if 'argument' in str(e):
                 msg = ("Wrong number of arguments supplied for action {0}."
@@ -288,6 +365,8 @@ class API(MethodView):
 
         self.session.commit()
 
+        for method in self._post_method.get('post', []):
+            method(self, retval)
         try:
             return jsonify(**retval)
         except (LeverException, KeyError, AttributeError):
@@ -307,6 +386,8 @@ class API(MethodView):
     def put(self):
         """ Updates an objects values """
         self.params = request.get_json(silent=True)
+        for method in self._pre_method.get('put', []):
+            method(self)
         if not self.params:
             raise LeverSyntaxError("To update, values must be specified")
         obj = self.get_obj()
@@ -317,33 +398,41 @@ class API(MethodView):
         for key, val in six.iteritems(self.params):
             current_app.logger.debug(
                 "Updating value for '{0}' to '{1}'".format(key, val))
-            assert obj.can('edit_' + key), "Can't edit key {0} on type {1}"\
+            assert self.can(obj, 'edit_' + key), "Can't edit key {0} on type {1}"\
                 .format(key, self.model.__name__)
             setattr(obj, key, val)
 
         self.session.commit()
+        retval = {'success': True}
+        for method in self._post_method.get('put', []):
+            method(self, retval)
 
-        return jsonify(success=True)
+        return jsonify(**retval)
 
     def delete(self):
         self.params = request.get_json(silent=True)
+        for method in self._pre_method.get('delete', []):
+            method(self)
         if not self.params:
             raise LeverSyntaxError("To delete, values must be specified")
 
         obj = self.get_obj()
         if not obj:
             raise LeverNotFound("Could not find any object to delete")
-        assert obj.can('delete'), "Can't delete that object"
+        assert self.can(obj, 'delete'), "Can't delete that object"
         self.session.delete(obj)
         self.session.commit()
 
-        return jsonify(success=True)
+        retval = {'success': True}
+        for method in self._post_method.get('delete', []):
+            method(self, retval)
+        return jsonify(**retval)
 
     def paginate(self, query=None):
         """ Sets limit and offset values on a query object based on arguments,
         and limited by class settings """
         if not query:
-            query = self.model.query
+            query = self.session.query(self.model)
         pg_size = self.params.get('pg_size')
         # don't do any pagination if we don't have a max page size and no
         # pagination is requested
@@ -359,7 +448,7 @@ class API(MethodView):
         """ Handles arguments __filter_by, __filter, and __order_by by
         modifying the query parameters before execution """
         if query is None:
-            query = self.model.query
+            query = self.session.query(self.model)
 
         filters = self.params.pop('__filter', None)
         try:
@@ -433,7 +522,7 @@ class API(MethodView):
         symfunc = cls.as_view(cls.__name__)
         mod.add_url_rule(url,
                          view_func=symfunc,
-                         methods=['GET', 'POST', 'PUT', 'PATCH', 'DELETE'])
+                         methods=['GET', 'POST', 'PUT', 'DELETE'])
 
 
 def get_joined(obj, join_prof="standard_join"):
